#ifndef _DEFS_H_
#define _DEFS_H_

#include "config.h"
#include "error.h"
#include "spinsdef.h"

#ifndef MAX_ITR		/* can be set now in config.h, Apr08 */
#define MAX_ITR 1000	/* maximum number of iterations, old=150 */
#endif

#if defined(__STRICT_ANSI__) || !(defined(__cplusplus) || defined(__GNUC__) || defined(__CPP__))
#define inline
#endif

/* please define VecType in config.h !!! Mar99 JS */
#ifndef VecType
#if defined(__cplusplus)
#define VecType      5  /* complex double as default for C++ */
#else
#define VecType      1	/* real double as default for C */
#endif
/*
    ToDo: reorder VecTypes bit0,1:0-3=flt8,flt16,flt32,flt64, bit2:4=cplx
                  SumTypes 0=flt64,1=flt80-128
    0=float,1=double,8=halfdouble=flt16(16bit,untested)
    4=complex-flt, 5=cplx-dbl, 12=cplx-flt16 (gcc, c++ or c99) for huge arrays
    2=long double (for sums)
 */
#endif

/*  plan: use short only for large vectors (less memory)!
 s=small l=long (2013-04-04)
         VecType:           0          1        2         !4      
 typical | typedef  |  short-  | normal- | long-  | real-version | old
 --------+----------+----------+---------+--------+--------------+--------
 array     scplx     sComplex    Complex   Complex  sdouble        mcplx,mzahl
 sums      lcplx      Complex    Complex  lComplex  ldouble        cplx,zahl
 rest       cplx      Complex    Complex   Complex   double        cplx,double

 [sl]cplx is identical to [sl]double in real mode (only real EVs)

*/

#if   (VecType & (1+8)) == 8          /* __GNUC__ 2.5.8 */
/* 2013-04-04 16bit fix point float removed,
 * because of bugs and bad  32bit tests
 * see ieee754.h union ieee754_float ... struct ...int ...:.. 32bit++
 * gcc-4.1.2 does not now about float16,
 *   ToDo:  does someone know about float16/binary16 defined in C99?
 */
typedef float16     sdouble; /* flt16 =  s1+e5+m10=16bits 3.3decdigits */
#elif (VecType & (1+8)) == 0 /* 32bits = a1+e8+m23  = 7.2decdigits */
typedef float       sdouble;    
#else                        /* 64bits = a1+e11+m52 = 15.9decdigits */
typedef double      sdouble;
/* ToDo: test 16bit fix vs. half-float s1+e5+m10 = 3.3decdigits */
#endif
#if (VecType & 2) == 2   /* extra accuracy for sums over lot of objects  */
/* gcc                      (default  80bits for x87,    s1+e15+m64) 19dec */
/* gcc -m96bit-long-double  (default  96bits for i386-32) */
/* gcc -m128bit-long-double (default 128bits for x86-64  s1+e15+m112) */
/* gcc -mlong-double-128    (default 128bits for x86-64  s1+e15+m112) 34dec */
typedef long double ldouble;  
#else
typedef double      ldouble;
#endif
#if VecType & 4
#define IFcplx(x) x       /* operations for COMPLEX only */
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
#define CplxInfo "c99 _Complex implementation"
/* gcc -std=c99 ... __STDC_VERSION__ 199901L _Complex ... */
/* using C99 complex type (faster) 2013-04-03 by A.Honecker */
/* tested: gcc -std=c99  VecType=  2013-04-04 gcc-4.1.2 test=OK */ 
#include <complex.h>
typedef complex double    cplx;
#if   (VecType & (1+8)) == 0          /* float for big arrays */
/* ToDo: C99: "double complex"(c99), not "complex double"(gcc) ! */
typedef complex float    scplx;
#elif (VecType & (1+8)) == 8          /* half float to store big arrays */
typedef complex float16  scplx; /* float16 _Complex = C99 s1+e5+m10=16bits */
#else
typedef complex double   scplx; /* default == 1 double for big arrays */
#endif
#if (VecType & 2) == 2          /* long double to store big arrays */
typedef complex long double  lcplx;
#else
typedef complex double       lcplx;
#endif
#define C(x,y)  ((x)+(y)*_Complex_I)
#define D(x)   ((double)creal(x))
#define Re(x)   creal(x)
#define Im(x)   cimag(x)
#define Conj(x) conj(x) 
#define arg(x) (atan2(cimag(x), creal(x)))   /* ToDo: carg(x) ??? */
#define polar(r,t) ((r) * (cos(t) + _Complex_I*sin(t)))
#define Norm(x) (creal(x)*creal(x)+cimag(x)*cimag(x))   /* r*r+i*i */
#elif defined(__STDC__) && defined(__GNUC__) && !defined(__cplusplus)
/* gcc-4.1.2 defaults to ISO C90 without __complex__ */
#define CplxInfo "c89+gnu __complex__ implementation"
/* GCC C89/C90 __complex__ float|double|long double (not ISO-C90) */
typedef __complex__ double   cplx;
#if   (VecType & (1+8)) == 0          /* float for big arrays */
typedef __complex__ float   scplx;
#elif (VecType & (1+8)) == 8          /* double to store big arrays */
typedef __complex__ float16 scplx;
#else
typedef __complex__ double  scplx;
#endif
#if (VecType & (2)) == 2          /* long double to store big arrays */
typedef __complex__ long double  lcplx;
#else 
typedef __complex__      double  lcplx; 
#endif
#define C(x,y)  ((x)+(y)*1i)
#define D(x)   ((double) __real__ (x))
#define Re(x)   (__real__ (x))
#define Im(x)   (__imag__ (x))
#define Conj(x) (~(x))                 /* ~(x) not on ISO C */
#define arg(x) (atan2(Im(x), Re(x)))   /* ToDo: carg(x) ??? */
#define polar(r,t) ((r) * (cos(t) + 1i*sin(t)))
#define Norm(x) (Re(x)*Re(x)+Im(x)*Im(x))   /* r*r+i*i */
#elif defined(__cplusplus)  /* using self defined class complex (preC99 c++ compilers) */
#define CplxInfo "slow(?) self written complex C++ class"
#include "complex.h"
typedef class  Complex  cplx; /* complex double      (compute) */ 
typedef class sComplex scplx; /* complex float       (storing) */
typedef class lComplex lcplx; /* complex long double (big sums) */
#define C(x,y)  Complex(x,y)
#define D(x)   ((double)real(x))
#define Re(x)   real(x)
#define Im(x)   imag(x)
#define Conj(x) conj(x)
#define Norm(x) norm(x)   /* r*r+i*i */
/* ToDo: may be its better to define I and use macros below */
#else  /* c++ or c99 */
#   error "use c++ OR add -std=c99 !!!"
#endif
#else  /* ! ( VecType & 4 ) == real, ToDo: define I as 0 ??? */
#define IFcplx(x)
typedef  double   cplx;
typedef sdouble  scplx; /* short for huge arrays */
typedef ldouble  lcplx; /* long  for huge summs */
#define C(x,y)  ((double)(x))
#define D(x)    ((double)(x))
#define Re(x)   (x)
#define Im(x)   0
#define Conj(x) (x)
#define Norm(x) norm(x)   /* not needed!? */
/* #define Norm(x) fsqr(x) */
#endif

#ifndef CplxInfo
#define CplxInfo ""
#endif
#if !defined(M_d2f) /* VecType & 8 needs special handling, see above */
//#define M_d2f(x) ((scplx)x)
//#define M_f2d(x) ( (cplx)x)
// ToDo: d2f+f2d is obsolete, before removing it, we test a while 2013-04-04
#define M_d2f(x) (x)
#define M_f2d(x) (x)
#endif

#include <limits.h>  /* __WORDSIZE (32,64) __LONG_LONG_MAX__ CHAR_BIT */
/* this values are fixed, but checked for sureness */
#ifndef S_short
#define S_short 2
#endif
#ifndef S_int
#define S_int 4
#endif
#ifndef S_long
#ifdef __WORDSIZE   /* linux */
#define S_long (__WORDSIZE/8)  /* 32 or 64 */
#elif defined(LONG_BIT) /* Tru64 */
#define S_long (LONG_BIT/8)    /* 32 or 64 */
#else  /* just guessing, check it later */
#define S_long (64/8)
#endif
#endif
/* long long is not ansi and therefore not used */
typedef unsigned long ULong;  /* already defined? -> where? Warnings */

#define Dsize sizeof(sdouble)
#define Lsize sizeof(long)
#define Csize sizeof(scplx)

/* using "\" in __TURBOC__ produces wrong line numbers!	*/
#define FF {}     	/* fflush(stdout) -> setvbuff(stdout,...)	*/

#ifdef sun
#undef sun                      /* no problems with 'SUN's (-Dsun) */
#endif

#include "array_m.h"		/* needs definition of scplx */

/* SLES9 forgets min max if array_m.h follows, cplx=dbl */
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))

#endif /* ifndef _DEFS_H_ */
